"""change paging_token to bigint

Revision ID: 4813b9c06c97
Revises: 19f853853773
Create Date: 2024-08-12 11:54:29.673037

"""
from typing import Sequence

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = '4813b9c06c97'
down_revision: str | None = '19f853853773'
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    with op.batch_alter_table('distribution_txs', schema=None) as batch_op:
        batch_op.alter_column('paging_token',
               existing_type=sa.INTEGER(),
               type_=sa.BigInteger(),
               existing_nullable=False)

    with op.batch_alter_table('minted_blocks', schema=None) as batch_op:
        batch_op.alter_column('paging_token',
               existing_type=sa.INTEGER(),
               type_=sa.BigInteger(),
               existing_nullable=False)

    with op.batch_alter_table('sinking_txs', schema=None) as batch_op:
        batch_op.alter_column('paging_token',
               existing_type=sa.INTEGER(),
               type_=sa.BigInteger(),
               existing_nullable=False)

    with op.batch_alter_table('test_distribution_txs', schema=None) as batch_op:
        batch_op.alter_column('paging_token',
               existing_type=sa.INTEGER(),
               type_=sa.BigInteger(),
               existing_nullable=False)

    with op.batch_alter_table('test_minted_blocks', schema=None) as batch_op:
        batch_op.alter_column('paging_token',
               existing_type=sa.INTEGER(),
               type_=sa.BigInteger(),
               existing_nullable=False)

    with op.batch_alter_table('test_sinking_txs', schema=None) as batch_op:
        batch_op.alter_column('paging_token',
               existing_type=sa.INTEGER(),
               type_=sa.BigInteger(),
               existing_nullable=False)


def downgrade() -> None:
    with op.batch_alter_table('test_sinking_txs', schema=None) as batch_op:
        batch_op.alter_column('paging_token',
               existing_type=sa.BigInteger(),
               type_=sa.INTEGER(),
               existing_nullable=False)

    with op.batch_alter_table('test_minted_blocks', schema=None) as batch_op:
        batch_op.alter_column('paging_token',
               existing_type=sa.BigInteger(),
               type_=sa.INTEGER(),
               existing_nullable=False)

    with op.batch_alter_table('test_distribution_txs', schema=None) as batch_op:
        batch_op.alter_column('paging_token',
               existing_type=sa.BigInteger(),
               type_=sa.INTEGER(),
               existing_nullable=False)

    with op.batch_alter_table('sinking_txs', schema=None) as batch_op:
        batch_op.alter_column('paging_token',
               existing_type=sa.BigInteger(),
               type_=sa.INTEGER(),
               existing_nullable=False)

    with op.batch_alter_table('minted_blocks', schema=None) as batch_op:
        batch_op.alter_column('paging_token',
               existing_type=sa.BigInteger(),
               type_=sa.INTEGER(),
               existing_nullable=False)

    with op.batch_alter_table('distribution_txs', schema=None) as batch_op:
        batch_op.alter_column('paging_token',
               existing_type=sa.BigInteger(),
               type_=sa.INTEGER(),
               existing_nullable=False)



"""
Load the relationships between sinking transactions and retirements into the DB.

These relationships are implicit in the source data. Retirement details include a string
representation of the included transaction hashes. This loader module iterates over retirement
records and stores the sink statuses of the corresponding transactions.

Author: Alex Olieman <https://keybase.io/alioli>
"""

from decimal import Decimal
from sqlalchemy import func, select

from sc_audit.db_schema.association import SinkStatus
from sc_audit.db_schema.retirement import Retirement
from sc_audit.db_schema.sink import SinkingTx
from sc_audit.session_manager import Session


def load_sink_statuses() -> int:
    """
    Load the retirement—sinking_tx associations to make the retirement status of sinking
    transactions explicit.
    """
    number_loaded = 0

    with Session.begin() as session:
        # select the retirements that have filled less than their VCU amount
        query = (
            select(Retirement)
            .outerjoin(Retirement.sink_statuses)
            .group_by(Retirement.certificate_id)
            .having(func.coalesce(func.sum(SinkStatus.amount_filled), 0) < Retirement.vcu_amount)
        )
        open_retirements = session.scalars(query).unique().all()
        for retirement in open_retirements:
            sink_statuses = create_sink_statuses(session, for_retirement=retirement)
            session.add_all(sink_statuses)
            number_loaded += len(sink_statuses)

    return number_loaded


def create_sink_statuses(session, for_retirement: Retirement) -> list[SinkStatus]:
    """
    For the given retirement, find which tx hashes do not yet have corresponding sink statuses.
    Keep track of the carbon amount remaining on the retirement while new statuses are created.
    """
    retirement = for_retirement
    existing_status_txs = {ss.sinking_tx_hash for ss in retirement.sink_statuses}
    existing_status_amount: Decimal = sum(
        (ss.amount_filled for ss in retirement.sink_statuses), start=Decimal()
    )
    new_sink_statuses: list[SinkStatus] = []
    for tx_hash in retirement.tx_hashes_from_details:
        if tx_hash in existing_status_txs:
            # only create new sink statuses
            continue

        sink_tx = session.get_one(SinkingTx, tx_hash)
        retirement_remaining = retirement.vcu_amount - existing_status_amount
        sink_remaining = sink_tx.carbon_amount - sink_tx.total_filled
        # status amount is the sink amount up to the remaining retirement amount
        amount_filled = min(sink_remaining, retirement_remaining)
        sink_status = SinkStatus(
            sinking_tx_hash=sink_tx.hash,
            certificate_id=retirement.certificate_id,
            amount_filled=amount_filled,
            finalized=((sink_tx.total_filled + amount_filled) == sink_tx.carbon_amount)
        )
        # indirect update of retirement_remaining and sink_tx.total_filled
        existing_status_amount += amount_filled
        sink_tx.statuses.append(sink_status)
        new_sink_statuses.append(sink_status)

    return new_sink_statuses



"""
SC Audit command line interface.

Author: Alex Olieman <https://keybase.io/alioli>
"""
import datetime as dt
from pathlib import Path
import shutil

import click

from sc_audit import migrations
from sc_audit.backup.download import download_compatible_dumps
from sc_audit.backup.dump import dump_table, get_table_names
from sc_audit.backup.restore import TABLE_LOADING_ORDER, get_table_row_counts, restore_all_tables, restore_table
from sc_audit.loader.__main__ import catch_up_from_sources
from sc_audit.loader.distribution_outflows import load_distribution_txs
from sc_audit.loader.get_latest import get_latest_attr
from sc_audit.loader.impact_projects import load_impact_projects
from sc_audit.loader.minted_blocks import load_minted_blocks
from sc_audit.loader.retirement_from_block import load_retirement_from_block
from sc_audit.loader.retirements import load_retirements
from sc_audit.loader.sink_events import load_sink_events
from sc_audit.loader.sink_status import load_sink_statuses
from sc_audit.loader.sinking_txs import load_sinking_txs
from sc_audit.sources.sink_events import MercuryError
from sc_audit.views.inventory import view_inventory
from sc_audit.views.retirement import view_retirements
from sc_audit.views.sink_status import view_sinking_txs
from sc_audit.views.utils import format_df


@click.group()
@click.version_option(package_name="sc_audit")
def cli():
    """CLI for Stellarcarbon Audit tool"""
    pass


@cli.command(name="catch-up")
@click.option("--fresh", is_flag=True, help="Don't bootstrap loading from a backup")
def db_catch_up(fresh: bool):
    """Let the DB catch up with the data sources"""
    if not fresh and not all(get_table_row_counts().values()):
        click.echo("Downloading database dumps...")
        download_dir = download_compatible_dumps()
        restore_all_tables(download_dir)
        shutil.rmtree(download_dir)
    
    catch_up_from_sources()

# VIEWS

@cli.group()
def view():
    """View the state of Stellarcarbon assets and transactions"""
    pass

view_format = click.Option(
    ["-f", "--format"],
    default="df",
    type=click.Choice(['df', 'csv', 'json'], case_sensitive=False),
    help="The output format for this view",
    show_default=True,
)

@view.command(name="inventory", params=[view_format])
@click.option("--omit-empty", is_flag=True, help="Don't show empty minted blocks")
@click.option(
    "--until-date", 
    type=click.DateTime(formats=["%Y-%m-%d"]), 
    help="Reconstruct the inventory on the given date"
)
def cli_view_inventory(omit_empty: bool, until_date: dt.datetime | None, format: str):
    """View Stellarcarbon's current or historical inventory of eco-credits"""
    until_dt_date = until_date.date() if until_date else None
    mbdf = view_inventory(omit_empty=omit_empty, until_date=until_dt_date)
    click.echo(format_df(mbdf, format=format))


@view.command(name="retirements", params=[view_format])
@click.option("--beneficiary", help="Only show retirements for the given beneficiary address")
@click.option(
    "--from-date", 
    type=click.DateTime(formats=["%Y-%m-%d"]), 
    help="Filter retirements that happened on or after the given date"
)
@click.option(
    "--before-date", 
    type=click.DateTime(formats=["%Y-%m-%d"]), 
    help="Filter retirements that happened before the given date"
)
@click.option("--project", type=int, help="Filter by impact project")
def cli_view_retirments(
        beneficiary: str | None, 
        from_date: dt.datetime | None,
        before_date: dt.datetime | None,
        project: int | None,
        format: str,
    ):
    """View finalized retirements and the attributes of the retired credits"""
    from_dt_date = from_date.date() if from_date else None
    before_dt_date = before_date.date() if before_date else None
    rtdf = view_retirements(
        for_beneficiary=beneficiary,
        from_date=from_dt_date,
        before_date=before_dt_date,
        project=project,
    )
    click.echo(format_df(rtdf, format=format))


@view.command(name="sink", params=[view_format])
@click.option("--funder", help="Only show transactions for the given funder address")
@click.option("--recipient", help="Only show transactions for the given recipient address")
@click.option(
    "--from-date", 
    type=click.DateTime(formats=["%Y-%m-%d"]), 
    help="Filter transactions that happened on or after the given date"
)
@click.option(
    "--before-date", 
    type=click.DateTime(formats=["%Y-%m-%d"]), 
    help="Filter transactions that happened before the given date"
)
@click.option("--finalized", type=bool, help="Filter by retirement status")
def cli_view_sink_status(
        funder: str | None,
        recipient: str | None, 
        from_date: dt.datetime | None,
        before_date: dt.datetime | None,
        finalized: bool | None,
        format: str,
    ):
    """View sinking transactions and their retirement status"""
    from_dt_date = from_date.date() if from_date else None
    before_dt_date = before_date.date() if before_date else None
    txdf = view_sinking_txs(
        for_funder=funder,
        for_recipient=recipient,
        from_date=from_dt_date,
        before_date=before_dt_date,
        finalized=finalized,
    )
    click.echo(format_df(txdf, format=format))

# LOADING

@cli.group()
def load():
    """Load data from their original sources"""
    pass


@load.command(name="impact-projects")
def db_load_impact_projects():
    """Load impact projects into the DB"""
    num_impact_projects = load_impact_projects()
    click.echo(f"Loaded {num_impact_projects} impact projects")


@load.command(name="minted-blocks")
def db_load_minted_blocks():
    """Load minted blocks into the DB"""
    mint_cursor = get_latest_attr('mint_tx')
    num_minting_txs = load_minted_blocks(cursor=mint_cursor) # type: ignore[arg-type]
    click.echo(f"Loaded {num_minting_txs} minted blocks")


@load.command(name="distribution-txs")
def db_load_distribution_outflows():
    """Load distribution outflows into the DB"""
    dist_cursor = get_latest_attr('dist_tx')
    num_distribution_txs = load_distribution_txs(cursor=dist_cursor) # type: ignore[arg-type]
    click.echo(f"Loaded {num_distribution_txs} distribution outflows")


@load.command(name="sinking-txs")
def db_load_sinking_txs():
    """Load sinking transactions into the DB"""
    sink_cursor: int = get_latest_attr('sink_tx') # type: ignore[return-value]
    num_sinking_txs = load_sinking_txs(cursor=sink_cursor)
    click.echo(f"Loaded {num_sinking_txs} sinking transactions")
    try:
        num_sink_events, _ = load_sink_events(cursor=sink_cursor)
        print(f"Loaded {num_sink_events} sink events")
    except MercuryError as exc:
        click.echo(f"Couldn't load sink events from Mercury")
        click.echo(repr(exc), err=True)


@load.command(name="retirements")
def db_load_retirements():
    """Load retirements into the DB"""
    retirement_date = get_latest_attr('retirement')
    num_retirements = load_retirements(from_date=retirement_date) # type: ignore[arg-type]
    click.echo(f"Loaded {num_retirements} retirements")


@load.command(name="associations")
def db_load_associations():
    """Load associations into the DB"""
    num_retirement_from_block = load_retirement_from_block()
    click.echo(f"Loaded {num_retirement_from_block} retirement-from-block associations")
    num_sink_statuses = load_sink_statuses()
    click.echo(f"Loaded {num_sink_statuses} sink status associations")

# MIGRATIONS

@cli.group()
def schema():
    """Database schema migration commands"""
    pass

@schema.command(name="current")
@click.option("-v", "--verbose", is_flag=True, default=False, help="Verbose mode")
def db_current_schema(verbose):
    """Display current database revision"""
    migrations.current(verbose)


@schema.command(name="upgrade")
@click.argument("revision", default="head")
def db_upgrade_schema(revision):
    """Upgrade to a later database revision"""
    migrations.upgrade(revision)


@schema.command(name="downgrade")
@click.argument("revision")
def db_downgrade_schema(revision):
    """Revert to a previous database revision. Specify "base" to reset the DB schema."""
    migrations.downgrade(revision)

# BACKUP

@cli.group()
def backup():
    """Database dump and restore commands"""
    pass


@backup.command(name="dump")
@click.argument("output_dir", type=click.Path(file_okay=False, dir_okay=True), required=True)
@click.option("-t", "--table", required=False, multiple=True, type=click.Choice(get_table_names()))
def db_dump_tables(output_dir, table: list[str]):
    """Dump the selected tables to ndjson files (default: all tables)"""
    tables = table  # ugly
    if not tables:
        tables = get_table_names()

    output = Path(output_dir)
    output.mkdir(exist_ok=True)
    for table_name in tables:
        file_path = output / f"{table_name}.ndjson"
        dump_table(db_model=table_name, output_path=file_path)
        click.echo(f"Wrote {table_name} to {file_path}")


@backup.command(name="restore")
@click.argument("input_dir", type=click.Path(file_okay=False, dir_okay=True), required=True)
@click.option("--replace", is_flag=True, help="Replace existing rows in tables with the backup")
@click.option("-t", "--table", required=False, multiple=True, help="The table name, see `dump`")
def db_restore_tables(input_dir, replace: bool, table: list[str]):
    """Restore the selected tables to the database (default: all files)"""
    selected_tables = table  # ugly
    input_dir = Path(input_dir)

    if selected_tables:
        for db_model in TABLE_LOADING_ORDER:
            table_name = db_model.__table__.name # type: ignore
            if table_name in selected_tables:
                dump_file = input_dir / f"{table_name}.ndjson"
                restore_table(dump_file, replace=replace)
    else:
        restore_all_tables(input_dir, replace=replace)

    


"""fix test table foreign keys

Revision ID: 0c2f50c4d517
Revises: 77ea86674a18
Create Date: 2025-06-18 12:10:02.734143

"""
from typing import Sequence

from alembic import context, op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = '0c2f50c4d517'
down_revision: str | None = '77ea86674a18'
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None

# get naming convention from the context
migration_context = context.get_context()
target_metadata = migration_context.opts.get("target_metadata")
assert target_metadata is not None, "Target metadata must be set in the migration context"
naming_convention = target_metadata.naming_convention


def upgrade() -> None:
    with op.batch_alter_table('test_minted_blocks', schema=None, naming_convention=naming_convention) as batch_op:
        batch_op.drop_constraint('test_minted_blocks_vcs_project_id_fkey', type_='foreignkey')
        batch_op.create_foreign_key(batch_op.f('test_minted_blocks_vcs_project_id_fkey'), 'test_vcs_projects', ['vcs_project_id'], ['id'])

    with op.batch_alter_table('test_retirement_from_block', schema=None, naming_convention=naming_convention) as batch_op:
        batch_op.drop_constraint('test_retirement_from_block_retirement_id_fkey', type_='foreignkey')
        batch_op.drop_constraint('test_retirement_from_block_block_hash_fkey', type_='foreignkey')
        batch_op.create_foreign_key(batch_op.f('test_retirement_from_block_retirement_id_fkey'), 'test_retirements', ['retirement_id'], ['certificate_id'])
        batch_op.create_foreign_key(batch_op.f('test_retirement_from_block_block_hash_fkey'), 'test_minted_blocks', ['block_hash'], ['serial_hash'])

    with op.batch_alter_table('test_retirements', schema=None, naming_convention=naming_convention) as batch_op:
        batch_op.drop_constraint('test_retirements_vcs_project_id_fkey', type_='foreignkey')
        batch_op.create_foreign_key(batch_op.f('test_retirements_vcs_project_id_fkey'), 'test_vcs_projects', ['vcs_project_id'], ['id'])

    with op.batch_alter_table('test_sink_status', schema=None, naming_convention=naming_convention) as batch_op:
        batch_op.drop_constraint('test_sink_status_certificate_id_fkey', type_='foreignkey')
        batch_op.drop_constraint('test_sink_status_sinking_tx_hash_fkey', type_='foreignkey')
        batch_op.create_foreign_key(batch_op.f('test_sink_status_certificate_id_fkey'), 'test_retirements', ['certificate_id'], ['certificate_id'])
        batch_op.create_foreign_key(batch_op.f('test_sink_status_sinking_tx_hash_fkey'), 'test_sinking_txs', ['sinking_tx_hash'], ['hash'])

    with op.batch_alter_table('test_sinking_txs', schema=None, naming_convention=naming_convention) as batch_op:
        batch_op.drop_constraint('test_sinking_txs_vcs_project_id_fkey', type_='foreignkey')
        batch_op.create_foreign_key(batch_op.f('test_sinking_txs_vcs_project_id_fkey'), 'test_vcs_projects', ['vcs_project_id'], ['id'])


def downgrade() -> None:
    with op.batch_alter_table('test_sinking_txs', schema=None, naming_convention=naming_convention) as batch_op:
        batch_op.drop_constraint(batch_op.f('test_sinking_txs_vcs_project_id_fkey'), type_='foreignkey')
        batch_op.create_foreign_key(None, 'vcs_projects', ['vcs_project_id'], ['id'])

    with op.batch_alter_table('test_sink_status', schema=None, naming_convention=naming_convention) as batch_op:
        batch_op.drop_constraint(batch_op.f('test_sink_status_sinking_tx_hash_fkey'), type_='foreignkey')
        batch_op.drop_constraint(batch_op.f('test_sink_status_certificate_id_fkey'), type_='foreignkey')
        batch_op.create_foreign_key(None, 'sinking_txs', ['sinking_tx_hash'], ['hash'])
        batch_op.create_foreign_key(None, 'retirements', ['certificate_id'], ['certificate_id'])

    with op.batch_alter_table('test_retirements', schema=None, naming_convention=naming_convention) as batch_op:
        batch_op.drop_constraint(batch_op.f('test_retirements_vcs_project_id_fkey'), type_='foreignkey')
        batch_op.create_foreign_key(None, 'vcs_projects', ['vcs_project_id'], ['id'])

    with op.batch_alter_table('test_retirement_from_block', schema=None, naming_convention=naming_convention) as batch_op:
        batch_op.drop_constraint(batch_op.f('test_retirement_from_block_block_hash_fkey'), type_='foreignkey')
        batch_op.drop_constraint(batch_op.f('test_retirement_from_block_retirement_id_fkey'), type_='foreignkey')
        batch_op.create_foreign_key(None, 'retirements', ['retirement_id'], ['certificate_id'])
        batch_op.create_foreign_key(None, 'minted_blocks', ['block_hash'], ['serial_hash'])

    with op.batch_alter_table('test_minted_blocks', schema=None, naming_convention=naming_convention) as batch_op:
        batch_op.drop_constraint(batch_op.f('test_minted_blocks_vcs_project_id_fkey'), type_='foreignkey')
        batch_op.create_foreign_key(None, 'vcs_projects', ['vcs_project_id'], ['id'])



"""
Load sink events from sorocarbon into the DB as SinkingTx records.

Author: Alex Olieman <https://keybase.io/alioli>
"""

from functools import lru_cache

from stellar_sdk.sep.toid import TOID

from sc_audit.config import settings
from sc_audit.db_schema.impact_project import get_vcs_project
from sc_audit.db_schema.sink import SinkingTx
from sc_audit.session_manager import Session
from sc_audit.sources.sink_events import get_sink_events


def load_sink_events(cursor: int=settings.FIRST_SINK_CURSOR) -> tuple[int, list[tuple[str, str]]]:
    """
    Load (all) sink events from Mercury Retroshades into the DB.

    To catch up with Retroshades, specify the cursor parameter to be the incremented
    paging token of the latest record present in the DB. Always get the latest cursor
    once, and then load both sinking transactions and sink events using that cursor.

    There is a real risk of desync and missing SinkingTxs if the cursor is refreshed
    between loading sinking transactions and sink events. The only reason to not load
    from both SinkingTx sources in unison is when you want to analyze one of them
    separately.
    """
    number_loaded = 0
    recipient_emails: dict[str, str] = {}

    with Session.begin() as session:
        last_ledger = 0
        # For each new ledger, reset tx_order to 2^16 to avoid collisions with SinkingTxs.
        # This is a pragmatic choice; ideally, the order should come from Retroshades.
        tx_order = first_tx_index = 2**16
        for sink_event in get_sink_events(cursor):
            if sink_event.email:
                recipient_emails[sink_event.recipient] = sink_event.email

            vcs_project_id = try_project_id(sink_event.project_id)

            # FIXME: the TOID should come from Retroshades if possible
            if sink_event.ledger == last_ledger:
                tx_order += 1
            else:
                tx_order = first_tx_index

            event_toid = TOID(sink_event.ledger, tx_order, 0)
            last_ledger = sink_event.ledger
            # TODO: support multiple events per transaction
            # We can support >1 event per tx by abandoning tx hash as PK, replacing it
            # with the TOID (paging_token). Not done yet, because it requires a data migration.

            memo = sink_event.memo_text or None
            session.add(
                SinkingTx(
                    hash=sink_event.transaction,
                    created_at=sink_event.created_at,
                    contract_id=sink_event.contract_id,
                    funder=sink_event.funder,
                    recipient=sink_event.recipient,
                    carbon_amount=sink_event.amount,
                    source_asset_code=settings.CARBON_ASSET.code,
                    source_asset_issuer=settings.CARBON_ISSUER_PUB,
                    source_asset_amount=sink_event.amount,
                    dest_asset_code=settings.CARBON_ASSET.code,
                    dest_asset_issuer=settings.CARBON_ISSUER_PUB,
                    dest_asset_amount=sink_event.amount,
                    vcs_project_id=vcs_project_id,
                    memo_type='text' if memo else 'none',
                    memo_value=memo,
                    paging_token=event_toid.to_int64(),
                )
            )
        
        number_loaded = len(session.new)

    return number_loaded, list(recipient_emails.items())


@lru_cache(maxsize=32)
def try_project_id(project_id: str) -> int:
    """
    Try to parse the project_id into a VCS project ID.
    If it fails, return 1360 for the default VCS project.

    TODO: replace fallback with AnyProject so we can be more flexible with retirements.
    """
    if project_id.startswith("VCS"):
        try:
            vcs_id = int(project_id[3:])
            vcs_project = get_vcs_project(vcs_id)
            if vcs_project:
                return vcs_project.id
        except ValueError:
            pass

    return 1360  # Default VCS project ID



"""
Fetch the current Carbon Pool inventory from the Verra Registry.

Author: Alex Olieman <https://keybase.io/alioli>
"""
import httpx
from parsel import Selector

from sc_audit.config import settings
from sc_audit.db_schema.mint import verra_carbon_pool
from sc_audit.sources.common import verra_default_headers


def get_carbon_pool_state():
    with httpx.Client(headers=verra_default_headers) as client:
        headers = {
            "accept": "text/html,application/xhtml+xml,application/xml;q=0.9",
            "referer": "https://www.stellarcarbon.io/",
        }
        query_params = {
            "r": 205,
            "idSubAccount": verra_carbon_pool.id,
        }
        resp: httpx.Response = client.get(
            url=str(settings.VERRA_REPORT_URL),
            headers=headers,
            params=query_params,
        )

    inventory_table = parse_inventory_table(resp.text)
    total_inventory_credits = sum(
        quantity_of_credits
        for batch in inventory_table
        if isinstance(quantity_of_credits := batch.get('quantity_of_credits'), int)
    )
    return {
        'total_inventory_credits': total_inventory_credits,
        'credit_batches': inventory_table,
    }


def cast_int(possible_int: str) -> int | str:
    try:
        return int(possible_int)
    except ValueError:
        return possible_int


def parse_inventory_table(report_html: str) -> list[dict[str, int | str]]:
    root = Selector(text=report_html)
    table_header = root.xpath('//td/font[starts-with(text(),"Serial Number")]/ancestor::tr[1]')
    table = table_header.xpath('parent::table')
    batch_rows = table.xpath('tr[position()>1]')

    column_names: list[str] = table_header.xpath('td/font/text()').getall()
    row_keys: list[str] = [
        name.strip().lower().replace(' ', '_').replace('-', '_').replace('/', '_')
        for name in column_names
    ]

    batch_items = [
        dict(zip(row_keys, [
            cast_int(val)
            for val in br.xpath('td/font//text()').getall()
        ]))
        for br in batch_rows
    ]
    return batch_items



"""
SQLAlchemy Declarative Dataclass setup.

Author: Alex Olieman <https://keybase.io/alioli>
"""
from dataclasses import fields
from decimal import Decimal
from typing import Annotated

from sqlalchemy import Dialect, MetaData, String
import sqlalchemy.types as types
from sqlalchemy.orm import mapped_column
from sqlalchemy.orm import DeclarativeBase, MappedAsDataclass

from sc_audit.config import settings


class HexBinary(types.TypeDecorator):
    """
    Converts between fixed-length bytes and their hexadecimal string representations.
    """
    impl = types.LargeBinary
    cache_ok = True

    def process_bind_param(self, value: str, dialect: Dialect) -> bytes:
        return bytes.fromhex(value)
    
    def process_result_value(self, value: bytes, dialect: Dialect) -> str:
        if value is None:
            # this can happen during joins or eager loading
            return ""
        
        return value.hex()


intpk = Annotated[int, mapped_column(primary_key=True)]
hashpk = Annotated[str, mapped_column(HexBinary(length=32), primary_key=True)]
strkey = Annotated[str, mapped_column(String(56))]
kgdecimal = Annotated[Decimal, mapped_column(types.DECIMAL(precision=21, scale=3))]
stroopdecimal = Annotated[Decimal, mapped_column(types.DECIMAL(precision=21, scale=7))]
bigint = Annotated[int, mapped_column(types.BigInteger())]


class ScBase(MappedAsDataclass, DeclarativeBase):
    metadata = MetaData(naming_convention={
        "ix": "idx_%(column_0_name)s",
        "uq": "uq_%(table_name)s_%(column_0_name)s",
        "ck": "ck_%(table_name)s_`%(constraint_name)s`",
        "fk": "%(table_name)s_%(column_0_name)s_fkey",
        "pk": "pk_%(table_name)s"
    })

    def __init_subclass__(cls) -> None:
        if settings.TABLE_PREFIX:
            # Check if the prefix has already been applied
            # unless the class is a test mapper class
            if (
                cls.__name__.startswith('Test') 
                or not cls.__tablename__.startswith(settings.TABLE_PREFIX)
            ):
                # Prefix the table name
                cls.__tablename__ = f"{settings.TABLE_PREFIX}_{cls.__tablename__}"

                for dfield in fields(cls):
                    field = getattr(cls, dfield.name, None)
                    if field and hasattr(field, 'foreign_keys'):
                        # Prefix all foreign key column references
                        for fk in field.foreign_keys:
                            if not fk._colspec.startswith(settings.TABLE_PREFIX):
                                fk._colspec = f"{settings.TABLE_PREFIX}_{fk._colspec}"
                            if fk.name and not fk.name.startswith(settings.TABLE_PREFIX):
                                fk.name = f"{settings.TABLE_PREFIX}_{fk.name}"

        super().__init_subclass__()


def create_test_mappers():
    """
    Create testnet versions of all registered models.
    These mappers are not needed at runtime, but they must exist when generating migrations.
    """
    models = [
        mapper.class_
        for mapper in ScBase.registry.mappers
    ]
    # TODO: generate indexes for test tables
    settings.TABLE_PREFIX = "test"
    for model in models:
        type(
            f'Test{model.__name__}', 
            model.__bases__, 
            {"__tablename__": model.__tablename__}
        )



from logging.config import fileConfig
from typing import Literal

from alembic import context
import sqlalchemy.types as types

from sc_audit.config import settings
from sc_audit.db_schema.base import HexBinary, ScBase, create_test_mappers
from sc_audit import session_manager


# create test mappers after all models have been registered
create_test_mappers()

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

# Interpret the config file for Python logging.
# This line sets up loggers basically.
if config.config_file_name is not None:
    fileConfig(config.config_file_name)

# add your model's MetaData object here
# for 'autogenerate' support
target_metadata = ScBase.metadata

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.


def render_item(type_, obj, autogen_context) -> str | Literal[False]:
    """Apply custom rendering for selected items."""

    if type_ == 'type' and isinstance(obj, HexBinary):
        # add import for this type
        autogen_context.imports.add("from sc_audit.db_schema.base import HexBinary")
        return f"{obj!r}"

    # default rendering for other objects
    return False


def my_compare_type(context, inspected_column,
            metadata_column, inspected_type, metadata_type) -> None | bool:
    """
    Return False if the metadata_type is the same as the inspected_type or None to allow the 
    default implementation to compare these types. A return value of True means the two types do 
    not match and should result in a type change operation.
    """
    if isinstance(metadata_type, HexBinary):
        if (
            isinstance(inspected_type, types.BINARY) 
            and inspected_type.length == metadata_type.length
        ):
            return False
        elif (
            isinstance(inspected_type, types.NUMERIC) 
            and inspected_type.precision == metadata_type.length
        ):
            return False
        else:
            return True
    elif isinstance(metadata_type, types.String):
        if (
            isinstance(inspected_type, types.String) 
            and inspected_type.length == metadata_type.length
        ):
            return False
        else:
            return True

    return None


def run_migrations_offline() -> None:
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    url = settings.DBAPI_URL
    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
        render_item=render_item,
        compare_type=my_compare_type,
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online() -> None:
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    connectable = session_manager.engine

    with connectable.connect() as connection:
        context.configure(
            connection=connection, 
            target_metadata=target_metadata, 
            render_item=render_item,
            compare_type=my_compare_type,
            render_as_batch=True,
        )

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



"""
View sinking transactions and their retirement status.

Transactions can be filtered by account, by date, and by their retirement status.

Author: Alex Olieman <https://keybase.io/alioli>
"""
import datetime as dt
from typing import Any, Literal

import pandas as pd
from sqlalchemy import Select, or_, select
from sqlalchemy.orm import contains_eager, selectinload

from sc_audit.db_schema.association import SinkStatus
from sc_audit.db_schema.sink import SinkingTx
from sc_audit.session_manager import Session


def view_sinking_txs(
        for_funder: str | None = None,
        for_recipient: str | None = None, 
        from_date: dt.date | None = None,
        before_date: dt.date | None = None,
        finalized: bool | None = None,
        cursor: int | None = None,
        limit: int | None = None,
        order: Literal['asc', 'desc'] = 'desc',
) -> pd.DataFrame:
    stx_query = construct_stx_query(for_funder, for_recipient, from_date, before_date, finalized)

    if order == 'asc':
        stx_query = stx_query.order_by(SinkingTx.created_at.asc())
        if cursor:
            stx_query = stx_query.where(SinkingTx.paging_token > cursor)
    if order == 'desc':
        stx_query = stx_query.order_by(SinkingTx.created_at.desc())
        if cursor:
            stx_query = stx_query.where(SinkingTx.paging_token < cursor)

    if limit:
        stx_query = stx_query.limit(limit)

    with Session.begin() as session:
        stx_records = session.scalars(stx_query).unique().all()
        txdf = pd.DataFrame.from_records(stx.as_dict() for stx in stx_records)

    return txdf


def construct_stx_query(
        for_funder: str | None,
        for_recipient: str | None, 
        from_date: dt.date | None,
        before_date: dt.date | None,
        finalized: bool | None,
) -> Select[tuple[SinkingTx]]:
    q_txs = select(SinkingTx).outerjoin(SinkStatus).options(contains_eager(SinkingTx.statuses))
    if for_funder:
        q_txs = q_txs.where(SinkingTx.funder == for_funder)

    if for_recipient:
        q_txs = q_txs.where(SinkingTx.recipient == for_recipient)

    if from_date:
        from_dt = dt.datetime(from_date.year, from_date.month, from_date.day)
        q_txs = q_txs.where(SinkingTx.created_at >= from_dt)

    if before_date:
        before_dt = dt.datetime(before_date.year, before_date.month, before_date.day)
        q_txs = q_txs.where(SinkingTx.created_at < before_dt)

    if finalized is False:
        # not_ and in_ do not work here because NULL and false are treated differently
        q_txs = q_txs.where(or_(SinkStatus.finalized == None, SinkStatus.finalized == False))
    elif finalized is True:
        q_txs = q_txs.where(SinkStatus.finalized == True)

    return q_txs


def get_sinking_tx(hash: str) -> dict[str, Any] | None:
    stx_query = (
        select(SinkingTx)
        .options(selectinload(SinkingTx.statuses))
        .where(SinkingTx.hash == hash)
    )
    with Session.begin() as session:
        stx = session.scalar(stx_query)
        stx_data = stx.as_dict() if stx else None

    return stx_data



"""
DB connection management.

Author: Alex Olieman <https://keybase.io/alioli>
"""
import sys

from sqlalchemy import Engine, create_engine, event
from sqlalchemy.engine.url import make_url
from sqlalchemy.orm import sessionmaker

from sc_audit.config import settings

print(f"Connecting to database {settings.DBAPI_URL}...", file=sys.stderr)

db_url = make_url(settings.DBAPI_URL)
connect_args = {}

if db_url.get_backend_name() == "sqlite":
    connect_args["check_same_thread"] = False

    @event.listens_for(Engine, "connect")
    def set_sqlite_pragma(dbapi_connection, connection_record):
        cursor = dbapi_connection.cursor()
        cursor.execute("PRAGMA foreign_keys=ON")
        cursor.close()

engine = create_engine(settings.DBAPI_URL, connect_args=connect_args)
Session = sessionmaker(engine)
